import inspect
import pickle
import typing

from . import helpers

VARARGS_VARIABLE = "_torchlayers_varargs_variable"
KWARGS_VARIABLE = "_torchlayers_kwargs_variable"

MODULE = "_torchlayers_infered_module"
MODULE_CLASS = "_torchlayers_infered_class_module"


def parse_arguments(
    init_arguments: typing.List[str], module
) -> typing.Tuple[typing.List[str], typing.Dict[str, typing.Any]]:
    """Parse init arguments.

    This function will:

        - create list of names which cannot be determined from non-instantiated module
        (e.g. `input_shape`) and assign them values (later used for non-instantiated __repr__)
        - remove type hints from argument to be inferred (the first one after `self`)
        and default value if provided
        - remove type hints from all `__init__` arguments and preserve their default
        arguments (custom exec won't be able to parse type hints without explicit imports)
        - remove `self` so it won't be later assigned to dynamically created non-instantiated
        module
        - WORKAROUND: add arguments related to recurrent neural networks to the uninferable group as those
        only have *args and **kwargs and are otherwise unparsable.


    Parameters
    ----------
    init_arguments : List[str]
        __init__ arguments gathered by `inspect.signature(cls).parameters`
    module : type
        `type` metaclass inheriting from `torch.nn.Module` and named like original
        class (without shape inference)

    Returns
    -------
    Tuple[List[str], Dict[str, Any]]
        First item are arguments used for dynamic __init__ creation of inferable
        module. Second item is dictionary where key is name of argument and value can be anything.
        Those are uninferable arguments (not present in inferable __init__)
        and used solely for inferable module's __repr__ creation.

    """

    def _add_rnn_uninferable(uninferable: typing.Dict, module) -> typing.Dict:
        if module.__name__ in ["RNN", "LSTM", "GRU"]:
            uninferable.update(
                {
                    "input_size": "?",
                    "num_layers": 1,
                    "bias": False,
                    "batch_first": False,
                    "dropout": 0.0,
                    "bidirectional": False,
                }
            )
        return uninferable

    def _add_infered_shape_variable_name(init_arguments: str) -> typing.Dict:
        first_argument = helpers.remove_right_side(
            helpers.remove_type_hint(init_arguments[1])
        )
        if not helpers.is_vararg(first_argument) and not helpers.is_kwarg(
            first_argument
        ):
            # Remove the argument to be inferred
            init_arguments.pop(1)
            return {first_argument: "?"}

        return {}

    uninferable_arguments = _add_rnn_uninferable(
        _add_infered_shape_variable_name(init_arguments), module
    )

    # Pop self from arguments always so it won't be class-assigned later
    init_arguments.pop(0)
    return (
        [helpers.remove_type_hint(argument) for argument in init_arguments],
        uninferable_arguments,
    )


def create_init(parsed_init_arguments) -> typing.Callable:
    """
    Dynamically create inferable module's `__init__`.

    Function has to be executed in `namespace` dictionary in order
    to build it from parsed strings.

    Function is string representation of constructor with proper inheritance
    of `torch.nn.Module`.

    If `*args` or `**kwargs` are encountered in original `__init__` they will
    be saved for later  in `VARARGS_VARIABLE` and `KWARGS_VARIABLE` respectively.

    They will be unpacked after all arguments during first `forward` call when
    module is instantiated.

    Parameters
    ----------
    parsed_init_arguments : List[str]
        __init__ arguments parsed by `parse_arguments` function.

    Returns
    -------
    Callable
        Function being __init__ of uninstantiated module.

    """
    namespace = {}

    joined_arguments = ", ".join(parsed_init_arguments)
    function = """def __init__(self, {}):""".format(joined_arguments)
    function += "\n super(type(self), self).__init__()\n"

    for argument in parsed_init_arguments:
        no_defaults = helpers.remove_right_side(argument)

        # Is vararg save for forward use
        if helpers.is_vararg(no_defaults):
            function += " self.{} = {}\n".format(
                VARARGS_VARIABLE, helpers.remove_vararg(no_defaults)
            )
        # Same for kwawrg
        elif helpers.is_kwarg(no_defaults):
            function += " self.{} = {}\n".format(
                KWARGS_VARIABLE, helpers.remove_kwarg(no_defaults)
            )
        # "Normal" variable, assign to self
        else:
            function += " self.{0} = {0}\n".format(no_defaults)

    exec(function, namespace)
    return namespace["__init__"]


def create_forward(
    module, module_class, parsed_init_arguments, inference_index: int
) -> typing.Callable:
    """
    Return forward function which instantiates module after first pass.

    Based on module type either `input.shape[1]` or `input.shape[2]` (for RNNs)
    will be gathered from input and used as first argument during `__init__`
    call of instantiated module.
    After that arguments saved in non-instantiated module will be passed, followed
    by `varargs` and `kwargs` (if any).

    Module will be fully usable after the first pass (has parameters, can be trained)
    but it is highly advised to use `torchlayers.build` to remove any non-instantiated
    module overlay.


    Parameters
    ----------
    module : str
        Name of variable where instantiated module will be saved. Usually equal to
        global variable `MODULE`
    module_class : torch.nn.Module class
        Name of variable where `__class__` of module to be instantiated is kept.
        Used to instantiate module only.
    parsed_init_arguments : List[str]
        __init__ arguments parsed by `parse_arguments` function. Used to get names
        of variables saved in non-instantiated module to be passed to `__init__`
        of to be instantiated module.

    Returns
    -------
    Callable
        Function being `forward` of uninstantiated module.

    """

    def _non_vararg_kwarg(arguments):
        return (
            helpers.remove_right_side(argument)
            for argument in arguments
            if not helpers.is_vararg(argument) and not helpers.is_kwarg(argument)
        )

    def forward(self, inputs, *args, **kwargs):
        infered_module = getattr(self, module, None)
        if infered_module is None:
            module_cls = getattr(self, module_class)
            # All arguments from non-instantiated module
            init_args = [
                getattr(self, argument)
                for argument in _non_vararg_kwarg(parsed_init_arguments)
                # List of varargs (if any)
            ] + list(getattr(self, VARARGS_VARIABLE, ()))
            # Kwargs to be unpacked (if any)
            init_kwargs = getattr(self, KWARGS_VARIABLE, {})

            self.add_module(
                module,
                module_cls(
                    # Shape to be inferred. Either 1 for all modules or 2 for RNNs
                    inputs.shape[inference_index],
                    *init_args,
                    **init_kwargs
                ),
            )

            infered_module = getattr(self, module)

        return infered_module(inputs, *args, **kwargs)

    return forward


def create_repr(module, **non_inferable_names) -> typing.Callable:
    """
    Uninstantiated module representation.

    Representation is gathered from saved variables names and values of module
    to be instantiated, saved `*args` variables (if any) and saved `**kwargs` variables.

    Additionally names not saved in module (e.g. `input_shape` to be inferred)
    are added at the beginning. Usually their value is "?", RNNs being and edge
    case with all their arguments in `non_inferable_names` group
    (and except the ones provided explicitly by user).


    Parameters
    ----------
    module : str
        Name of variable where instantiated module will be saved. Usually equal to
        global variable `MODULE`
    **non_inferable_names : Dict[str, Any]
        Non-inferable names and their respective values of the module

    Returns
    -------
    Callable
        Function being `__repr__` of uninstantiated module.

    """

    def __repr__(self) -> str:
        infered_module = getattr(self, module, None)
        if infered_module is None:
            parameters = ", ".join(
                argument_representation
                for argument_representation in helpers.create_vars(
                    self, non_inferable_names, VARARGS_VARIABLE, KWARGS_VARIABLE
                )
            )
            return "{}({})".format(type(self).__name__, parameters)

        return repr(infered_module)

    return __repr__


def create_getattr(module) -> typing.Callable:
    """
    Create __getattr__ of uninstantiated module.

    Will return values from `uninstantiated` network if exist, if not
    will check it's `instantiated` network (if it exists), otherwise
    return `NoAttributeError`.

    Can be considered proxy passing user calls `module` after instantiation.

    Parameters
    ----------
    module : str
        Name of variable where instantiated module will be saved. Usually equal to
        global variable `MODULE`

    Returns
    -------
    Callable
        __getattr__ function
    """

    def __getattr__(self, name) -> str:
        if name == module:
            return super(type(self), self).__getattr__(name)
        return getattr(getattr(self, module), name)

    return __getattr__


# FIXED FOR PyTorch 1.4.0, 1.2.0 should work fine as well although it may throw warnings

# For warning regarding inability to find source code
# https://github.com/pytorch/pytorch/blob/master/torch/_utils_internal.py#L44

# getsourcefile
# https://github.com/python/cpython/blob/master/Lib/inspect.py#L692
# getfile
# https://github.com/python/cpython/blob/master/Lib/inspect.py#L654
# Simulate self.__module__.__file__ variable appropriately

# getsourcelines
# https://github.com/python/cpython/blob/master/Lib/inspect.py#L958


def create_reduce(module, parsed_init_arguments):
    """
    Create __reduce__ of uninstantiated module.

    This is one of core functionalities. Using custom `__reduce__` modules
    are reduced to their instantiated (after shape inference) versions.

    Due to this operation and `torchlayers.build` zero-overhead of shape
    inference is provided and support of `torchscript`.

    If the module was not instantiated `ValueError` error is thrown.

    Parameters
    ----------
    module : str
        Name of variable where instantiated module will be saved. Usually equal to
        global variable `MODULE`

    parsed_init_arguments : List[str]
        __init__ arguments parsed by `parse_arguments` function. Used to get names
        of variables saved in non-instantiated module to be passed to `__init__`
        of to be instantiated module.

    Returns
    -------
    Callable
        __getattr__ function
    """

    def __reduce__(self):
        infered_module = getattr(self, module, None)
        if infered_module is None:
            raise ValueError(
                "Model cannot be pickled as it wasn't instantiated. Pass example input through this module."
            )

        custom_reduce = getattr(infered_module, "__reduce__", None)
        if custom_reduce is None:
            raise ValueError(
                "Infered module does not have a __reduce__ method. Does it inherit from torch.nn.Module?"
            )
        return custom_reduce()

    return __reduce__



import typing

import torch

from . import _dev_utils


class InferDimension(torch.nn.Module):
    """Infer dimensionality of module from input using dispatcher.

    Users can pass provide their own modules to infer dimensionality
    from input tensor by inheriting from this module and providing
    `super().__init__()` with `dispatcher` method::


        import torchlayers as tl

        class BatchNorm(tl.InferDimension):
            def __init__(
                self,
                num_features: int,
                eps: float = 1e-05,
                momentum: float = 0.1,
                affine: bool = True,
                track_running_stats: bool = True,
            ):
                super().__init__(
                    dispatcher={
                        # 5 dimensional tensor -> create torch.nn.BatchNorm3d
                        5: torch.nn.BatchNorm3d,
                        # 4 dimensional tensor -> create torch.nn.BatchNorm2d
                        4: torch.nn.BatchNorm2d,
                        3: torch.nn.BatchNorm1d,
                        2: torch.nn.BatchNorm1d,
                    },
                    num_features=num_features,
                    eps=eps,
                    momentum=momentum,
                    affine=affine,
                    track_running_stats=track_running_stats,
                )

    All dimension-agnostic modules in `torchlayers` are created this way.
    This class can also be mixed with `torchlayers.infer` for dimensionality
    and shape inference in one.

    This class works correctly with `torchlayers.build` and other provided
    functionalities.

    Parameters
    ----------
    dispatcher: Dict[int, torch.nn.Module]
        Key should be length of input's tensor shape. Value should be a `torch.nn.Module`
        to be used for the dimensionality.
    initializer: Callable[[torch.nn.Module, torch.Tensor, **kwargs], torch.nn.Module], optional
        How to initialize dispatched module. Can be used to modify it's creation.
        First argument - dispatched module class, second - input tensor, **kwargs are
        arguments to use for module initialization. Should return module's instance.
        By default dispatched module initialized with **kwargs is returned.
    **kwargs:
        Arguments used to initialize dispatched module

    """

    def __init__(
        self,
        dispatcher: typing.Dict[int, torch.nn.Module],
        initializer: typing.Callable = None,
        **kwargs,
    ):
        super().__init__()

        self._dispatcher = dispatcher
        self._inner_module_name = "_inner_module"

        if initializer is None:
            self._initializer = lambda dispatched_class, _, **kwargs: dispatched_class(
                **kwargs
            )
        else:
            self._initializer = initializer

        for key, value in kwargs.items():
            setattr(self, key, value)

        self._noninferable_attributes = [key for key in kwargs]
        self._repr = _dev_utils.infer.create_repr(self._inner_module_name, **kwargs)
        self._reduce = _dev_utils.infer.create_reduce(
            self._inner_module_name, self._noninferable_attributes
        )

    def __repr__(self):
        return self._repr(self)

    def __reduce__(self):
        return self._reduce(self)

    def forward(self, inputs):
        module = getattr(self, self._inner_module_name, None)
        if module is None:
            dimensionality = len(inputs.shape)
            dispatched_class = self._dispatcher.get(dimensionality)
            if dispatched_class is None:
                dispatched_class = self._dispatcher.get("*")
                if dispatched_class is None:
                    raise ValueError(
                        "{} could not be inferred from shape. Got tensor of dimensionality: {} but only {} are allowed".format(
                            self._module_name,
                            dimensionality,
                            list(self._dispatcher.keys()),
                        )
                    )

            self.add_module(
                self._inner_module_name,
                self._initializer(
                    dispatched_class,
                    inputs,
                    **{
                        key: getattr(self, key) for key in self._noninferable_attributes
                    },
                ),
            )

        return getattr(self, self._inner_module_name)(inputs)



import collections
import itertools
import typing


def is_kwarg(argument: str) -> bool:
    """`True` if argument name is `**kwargs`"""
    return "**" in argument


def is_vararg(argument: str) -> bool:
    """`True` if argument name is `*args`"""
    return "*" in argument and "**" not in argument


def remove_vararg(argument: str) -> str:
    """`True` if argument name is `*args`"""
    return argument.split("*")[1]


def remove_kwarg(argument: str) -> str:
    """Removes `**` in kwargs so those can be assigned in dynamic `__init__` creation"""
    return argument.split("**")[1]


def remove_right_side(argument: str) -> str:
    """Removes default arguments so their names can be used as values or names of variables."""
    return argument.split("=")[0]


def remove_type_hint(argument: str) -> str:
    """Removes any Python type hints.

    Those are incompatible with `exec` during dynamic `__init__` creation
    (at least not without major workarounds).

    Default values (right hand side) is preserved if exists.

    """
    splitted_on_type_hint = argument.split(":")
    no_type_hint = splitted_on_type_hint[0]
    if len(splitted_on_type_hint) > 1:
        splitted_on_default = splitted_on_type_hint[1].split("=")
        if len(splitted_on_default) > 1:
            no_type_hint += "={}".format(splitted_on_default[1])
    return no_type_hint


def create_vars(
    self,
    non_inferable_names: typing.Dict[str, typing.Any],
    varargs_variable: str,
    kwargs_variable: str,
) -> typing.List[str]:
    """
    Create list of arguments for uninstantiated `__repr__` module.

    Parameters
    ----------
    **non_inferable_names : Dict[str, Any]
        Non-inferable names and their respective values of the module
    varargs_variable : str
        Name of variable possibly holding varargs for module's __init__.
    kwargs_variable : str
        Name of variable possibly holding kwargs for module's __init__.

    Returns
    -------
    typing.List[str]:
        List of strings formatted in the manner argument=value to display
        for uninstantiated module.
    """
    dictionary = {**non_inferable_names, **collections.OrderedDict(vars(self))}

    args = [
        "{}={}".format(key, value)
        for key, value in dictionary.items()
        if not key.startswith("_") and key != "training"
    ]
    varargs = getattr(self, varargs_variable, None)
    if varargs is not None:
        args += [str(var) for var in varargs]

    kwargs = getattr(self, kwargs_variable, None)
    if kwargs is not None:
        args += ["{}={}".format(key, value) for key, value in kwargs.items()]

    return args



import collections

import torch

from . import module


class InstanceNorm(module.InferDimension):
    """Apply Instance Normalization over inferred dimension (3D up to 5D).

    Based on input shape it either creates 1D, 2D or 3D instance normalization for inputs of shape
    3D, 4D, 5D respectively (including batch as first dimension).

    Otherwise works like standard PyTorch's `InstanceNorm <https://pytorch.org/docs/stable/nn.html#torch.nn.InstanceNorm1d>`__

    Parameters
    ----------
    num_features : int
        :math:`C` (number of channels in input) from an expected input.
        Can be number of outputs of previous linear layer as well
    eps : float, optional
        Value added to the denominator for numerical stability.
        Default: `1e-5`
    momentum : float, optional
        Value used for the `running_mean` and `running_var`
        computation. Default: `0.1`
    affine : bool, optional
        If ``True``, this module has learnable affine parameters, initialized just like in batch normalization.
        Default: ``False``
    track_running_stats : bool, optional
        If ``True``, this module tracks the running mean and variance,
        and when set to ``False``, this module does not track such statistics and always uses batch
        statistics in both training and eval modes.
        Default: ``False``

    """

    def __init__(
        self,
        num_features: int,
        eps: float = 1e-05,
        momentum: float = 0.1,
        affine: bool = False,
        track_running_stats: bool = False,
    ):
        super().__init__(
            dispatcher={
                5: torch.nn.InstanceNorm3d,
                4: torch.nn.InstanceNorm2d,
                3: torch.nn.InstanceNorm1d,
            },
            num_features=num_features,
            eps=eps,
            momentum=momentum,
            affine=affine,
            track_running_stats=track_running_stats,
        )


class BatchNorm(module.InferDimension):
    """Apply Batch Normalization over inferred dimension (2D up to 5D).

    Based on input shape it either creates `1D`, `2D` or `3D` batch normalization for inputs of shape
    `2D/3D`, `4D`, `5D` respectively (including batch as first dimension).

    Otherwise works like standard PyTorch's `BatchNorm <https://pytorch.org/docs/stable/nn.html#batchnorm1d>`__.

    Parameters
    ----------
    num_features : int
        :math:`C` (number of channels in input) from an expected input.
        Can be number of outputs of previous linear layer as well
    eps : float, optional
        Value added to the denominator for numerical stability.
        Default: `1e-5`
    momentum : float, optional
        Value used for the `running_mean` and `running_var`
        computation. Can be set to ``None`` for cumulative moving average
        (i.e. simple average). Default: `0.1`
    affine : bool, optional
        If ``True``, this module has learnable affine parameters.
        Default: ``True``
    track_running_stats : bool, optional
        If ``True``, this module tracks the running mean and variance,
        and when set to ``False``, this module does not track such statistics and always uses batch
        statistics in both training and eval modes.
        Default: ``True``

    """

    def __init__(
        self,
        num_features: int,
        eps: float = 1e-05,
        momentum: float = 0.1,
        affine: bool = True,
        track_running_stats: bool = True,
    ):
        super().__init__(
            dispatcher={
                5: torch.nn.BatchNorm3d,
                4: torch.nn.BatchNorm2d,
                3: torch.nn.BatchNorm1d,
                2: torch.nn.BatchNorm1d,
            },
            num_features=num_features,
            eps=eps,
            momentum=momentum,
            affine=affine,
            track_running_stats=track_running_stats,
        )


# Arguments had the wrong order unfortunately
class GroupNorm(torch.nn.GroupNorm):
    """Apply Group Normalization over a mini-batch of inputs.

    Works exactly like PyTorch's counterpart, but `num_channels` is used as first argument
    so it can be inferred during first forward pass.

    Parameters
    ----------
    num_channels : int
        Number of channels expected in input
    num_groups : int
        Number of groups to separate the channels into
    eps : float, optional
        Value added to the denominator for numerical stability.
        Default: `1e-5`
    affine : bool, optional
        If ``True``, this module has learnable affine parameters.
        Default: ``True``

    """

    def __init__(
        self,
        num_channels: int,
        num_groups: int,
        eps: float = 1e-05,
        affine: bool = True,
    ):
        super().__init__(
            num_groups=num_groups, num_channels=num_channels, eps=eps, affine=affine,
        )
